"""Options for slices-cli."""

OPTIONS_WITHOUT_ARGS = [
    "--version",
    "--help",
    "-h",
    "--no-upgrade-check",
    "-v",
    "-vv",
    "-vvv",
]

OPTIONS_WITH_ARGS = []

OPTIONS_WITHOUT_ARGS__HELP = {
    "--version": "Show version and exit.",
    "--help": "Show help and exit.",
    "-h": "Show help and exit.",
    "--no-upgrade-check": "Do not check if new CLI versions are available..",
}

OPTIONS_WITH_ARGS__HELP = {}
